/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 2; tab-width: 2 -*- */
/* gdome-xml-node.c
 *
 * Copyright (C) 1999 Raph Levien <raph@acm.org>
 * Copyright (C) 2000 Mathieu Lacage <mathieu@gnu.org>
 * Copyright (C) 2000 Anders Carlsson <andersca@gnu.org>
 * Copyright (C) 2001 Paolo Casarini <paolo@casarini.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

#include <sys/time.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include "gdome.h"
#include "gdome-util.h"
#include <tree.h>
#include <parser.h>
#include <xmlmemory.h>
#include <hash.h>
#include "gdome-xml-util.h"
#include "gdome-xml-xmlutil.h"
#include "gdome-xml-xmldtdutil.h"
#include "gdome-xml-str.h"
#include "gdome-xml-event.h"
#include "gdome-xml-eventl.h"
#include "gdome-xml-mevent.h"
#include "gdome-xml-nodel.h"
#include "gdome-xml-node.h"
#include "gdome-xml-document.h"
#include "gdome-xml-element.h"
#include "gdome-xml-attribute.h"
#include "gdome-xml-cdata.h"
#include "gdome-xml-text.h"
#include "gdome-xml-cdatas.h"
#include "gdome-xml-comment.h"
#include "gdome-xml-entity.h"
#include "gdome-xml-pi.h"
#include "gdome-xml-notation.h"
#include "gdome-xml-documentt.h"
#include "gdome-xml-documentf.h"

/* indexed by node type 
	 first one is NULL because there 
   is no node type number 0 
   cf: spec p25 (DOM 1). interface Node.
*/
const void * gdome_xml_vtabs[] = {
	NULL, 
	&gdome_xml_el_vtab,
	&gdome_xml_a_vtab,
	&gdome_xml_t_vtab,
	&gdome_xml_cds_vtab,
	NULL,	
	&gdome_xml_ent_vtab,
	&gdome_xml_pi_vtab,
	&gdome_xml_c_vtab,
	&gdome_xml_doc_vtab, 
	&gdome_xml_dt_vtab,
	&gdome_xml_df_vtab,
	&gdome_xml_not_vtab,
	&gdome_xml_doc_vtab,
  &gdome_xml_dt_vtab,
  NULL,
  NULL,
  &gdome_xml_ent_vtab
};

/***************
 * Virtual Table
 **************/
const GdomeNodeVtab gdome_xml_n_vtab = 
{
	gdome_xml_n_ref,
	gdome_xml_n_unref,
	gdome_xml_n_query_interface,
	gdome_xml_n_nodeName,
	gdome_xml_n_nodeValue,
	gdome_xml_n_set_nodeValue,
	gdome_xml_n_nodeType,
	gdome_xml_n_parentNode,
	gdome_xml_n_childNodes,
	gdome_xml_n_firstChild,
	gdome_xml_n_lastChild,
	gdome_xml_n_previousSibling,
	gdome_xml_n_nextSibling,
	gdome_xml_n_attributes,
	gdome_xml_n_ownerDocument,
	gdome_xml_n_insertBefore,
	gdome_xml_n_replaceChild,
	gdome_xml_n_removeChild,
	gdome_xml_n_appendChild,
	gdome_xml_n_hasChildNodes,
	gdome_xml_n_cloneNode,
	gdome_xml_n_normalize,
	gdome_xml_n_isSupported,
	gdome_xml_n_namespaceURI,
	gdome_xml_n_prefix,
	gdome_xml_n_set_prefix,
  gdome_xml_n_localName,
  gdome_xml_n_hasAttributes,
	gdome_xml_n_addEventListener,
	gdome_xml_n_removeEventListener,
	gdome_xml_n_dispatchEvent
};


/* function to get the xmlNode from a GdomeNode */
xmlNode *
gdome_xml_n_get_xmlNode (GdomeNode *node)
{
	Gdome_xml_Node *xmlnode;

	xmlnode = (Gdome_xml_Node *)node;

	return xmlnode->n;
}

/***********************************************************
 ***** Constructor and Reference Count Implementation ******
 ***********************************************************/

/**
 * gdome_xml_n_mkref:
 * @n  The libxml node of which a Gdome reference is wanted
 *
 * Constructor of the Node structure.
 * Returns the reference to the Gdome node wanted.
 */
GdomeNode *
gdome_xml_n_mkref (xmlNode *n) {
	Gdome_xml_Node *result = NULL;

	if (n == NULL)
		return NULL;
	if (n->_private) {
		result = (Gdome_xml_Node *)n->_private;
		result->refcnt++;
	} else {
		xmlElementType type = n->type;

		switch (type) {
		case XML_ELEMENT_NODE:
		case XML_TEXT_NODE:
		case XML_CDATA_SECTION_NODE:
		case XML_ENTITY_REF_NODE:
		case XML_PI_NODE:
		case XML_COMMENT_NODE:
		case XML_NOTATION_NODE:
		case XML_DOCUMENT_NODE:
		case XML_DOCUMENT_FRAG_NODE:
		case XML_HTML_DOCUMENT_NODE:
		case XML_ATTRIBUTE_NODE:
		case XML_ENTITY_DECL: /*libxml use this instead of XML_ENTITY_NODE*/
			result = g_new (Gdome_xml_Node, 1);
			result->refcnt = 1;
			result->super.vtab = (GdomeNodeVtab *)gdome_xml_vtabs[n->type];
      result->n = n;
      if (type == XML_NOTATION_NODE || type == XML_ENTITY_NODE || type == XML_ENTITY_DECL)
        result->accessType = GDOME_READONLY_NODE;
      else
        result->accessType = GDOME_READWRITE_NODE;
			result->ll = NULL;
			n->_private = result;
			break;
		default:
			g_warning ("gdome_xml_n_mkref: invalid node type");
			break;
		}
	}
	return (GdomeNode *)result;
}

/**
 * gdome_xml_ref:
 * @self  Node Object ref
 * @exc  Exception Object ref
 *
 * Increase the reference count of the specified Node.
 */
void
gdome_xml_n_ref (GdomeNode *self, GdomeException *exc) {
	Gdome_xml_Node *priv = (Gdome_xml_Node *)self;

	if (priv == NULL)
		return;
  g_return_if_fail (GDOME_XML_IS_N (priv));
  g_return_if_fail (exc != NULL);

	priv->refcnt++;
}

/**
 * gdome_xml_n_unref:
 * @self  Node Object ref
 * @exc  Exception Object ref
 *
 * Decrease the reference count of the specified Node. Free the Node structure
 * if the Node will have zero reference.
 */
void
gdome_xml_n_unref (GdomeNode *self, GdomeException *exc) {
	Gdome_xml_Node *priv = (Gdome_xml_Node *)self;

	if (priv == NULL)
		return;
	g_return_if_fail (GDOME_XML_IS_N (priv));
  g_return_if_fail (exc != NULL);

	priv->refcnt--;

	if (priv->refcnt == 0) {
		xmlNode *n = (xmlNode *)priv->n;

		n->_private = NULL;
		g_free (self);
	}
}

void *
gdome_xml_n_query_interface (GdomeNode *self, const char *interface, GdomeException *exc)
{
	Gdome_xml_Node *priv = (Gdome_xml_Node *)self;

	if (priv == NULL)
		return NULL;
	if (interface == NULL)
		return NULL;
	if (exc == NULL)
		return NULL;
	g_return_val_if_fail (GDOME_XML_IS_N (priv), NULL);

	if (!strcmp (interface, "Node") || !strcmp (interface, "EventTarget"))
		{
			priv->refcnt++;
			return self;
		}
	else
		return NULL;
}

/**************************************
 ***** Attributes implementation ******
 **************************************/

/**
 * gdome_xml_n_attributes:
 * @self:  Node Object ref
 * @exc:  Exception Object ref
 *
 * Returns a NamedNodeMap containing the attributes of this node (if it is an
 * Element) or null otherwise.
 */
GdomeNamedNodeMap *
gdome_xml_n_attributes (GdomeNode *self, GdomeException *exc)
{
	g_return_val_if_fail (self != NULL, NULL);
	g_return_val_if_fail (GDOME_XML_IS_N (self), NULL);
	g_return_val_if_fail (exc != NULL, NULL);

	if (GDOME_XML_IS_EL (self))
		return gdome_el_attributes((GdomeElement *)self, exc);

	return NULL;
}

/**
 * gdome_xml_n_childNodes:
 * @self:  Node Object ref
 * @exc:  Exception Object ref
 *
 * Returns a NodeList that contains all children of this node. If there are no
 * children, this is a NodeList containing no nodes.
 */
GdomeNodeList *
gdome_xml_n_childNodes (GdomeNode *self, GdomeException *exc)
{
	g_return_val_if_fail (self != NULL, NULL);
	g_return_val_if_fail (GDOME_XML_IS_N (self), NULL);
	g_return_val_if_fail (exc != NULL, NULL);

  return gdome_xml_nl_mkref ((GdomeNode*)self, NULL, NULL, GDOME_READONLY_NODE);
}

/**
 * gdome_xml_n_firstChild:
 * @self:  Node Object ref
 * @exc:  Exception Object ref
 * 
 * Returns the first child (readonly) of this node. If there is no such node,
 * this returns NULL.
 */
GdomeNode *
gdome_xml_n_firstChild (GdomeNode *self, GdomeException *exc)
{
	Gdome_xml_Node *priv = (Gdome_xml_Node *)self;

	g_return_val_if_fail (priv != NULL, NULL);
	g_return_val_if_fail (GDOME_XML_IS_N (priv), NULL);
	g_return_val_if_fail (exc != NULL, NULL);

	return gdome_xml_n_mkref (gdome_xmlGetFirstChild (priv->n));
}

/**
 * gdome_xml_n_lastChild:
 * @self:  Node Object ref
 * @exc:  Exception Object ref
 *
 * Returns the last child (readonly) of this node. If there is no such node,
 * this returns NULL.
 */
GdomeNode *
gdome_xml_n_lastChild (GdomeNode *self, GdomeException *exc)
{
	Gdome_xml_Node *priv = (Gdome_xml_Node *)self;

	g_return_val_if_fail (priv != NULL, NULL);
	g_return_val_if_fail (GDOME_XML_IS_N (priv), NULL);
	g_return_val_if_fail (exc != NULL, NULL);

	return gdome_xml_n_mkref (gdome_xmlGetLastChild (priv->n));
}

/**
 * gdome_xml_n_localName:
 * @self:  Node Objects ref
 * @exc:  Exception Object ref
 *
 * Returns the local part of the qualified name of this node.
 * For nodes of any type other than ELEMENT_NODE and ATTRIBUTE_NODE and nodes
 * created with a DOM Level 1 method, this is always NULL.
 */
GdomeDOMString *
gdome_xml_n_localName (GdomeNode *self, GdomeException *exc)
{
	Gdome_xml_Node *priv = (Gdome_xml_Node *)self;

	g_return_val_if_fail (priv != NULL, NULL);
	g_return_val_if_fail (GDOME_XML_IS_N (priv), NULL);
	g_return_val_if_fail (exc != NULL, NULL);

	if((gdome_xmlGetType(priv->n) == XML_ELEMENT_NODE ||
			gdome_xmlGetType(priv->n) == XML_ATTRIBUTE_NODE) &&
		 gdome_xmlGetNs(priv->n) != NULL)
		return gdome_xml_str_mkref_dup ((gchar *)gdome_xmlGetName (priv->n));
	else
		return NULL;
}

/**
 * gdome_xml_n_namespaceURI:
 * @self:  Node Objects ref
 * @exc:  Exception Object ref
 *
 * Returns the namespace URI of this node, or NULL if it is unspecified.
 * For nodes of any type other than ELEMENT_NODE and ATTRIBUTE_NODE and nodes
 * created with a DOM Level 1 method, this is always NULL.
 */
GdomeDOMString *
gdome_xml_n_namespaceURI (GdomeNode *self, GdomeException *exc)
{
	Gdome_xml_Node *priv = (Gdome_xml_Node *)self;

	g_return_val_if_fail (priv != NULL, NULL);
	g_return_val_if_fail (GDOME_XML_IS_N (priv), NULL);
	g_return_val_if_fail (exc != NULL, NULL);

	if((gdome_xmlGetType(priv->n) == XML_ELEMENT_NODE ||
			gdome_xmlGetType(priv->n) == XML_ATTRIBUTE_NODE) &&
		 gdome_xmlGetNs(priv->n) != NULL)
		return gdome_xml_str_mkref_dup ((gchar *)gdome_xmlGetNsURI (priv->n));
	else
		return NULL;
}

/**
 * gdome_xml_n_nextSibling:
 * @self:  Node Object ref
 * @exc:  Exception Object ref
 *
 * The node (readonly) immediately following this node. If there is no such
 * node, this returns NULL.
 */
GdomeNode *
gdome_xml_n_nextSibling (GdomeNode *self, GdomeException *exc)
{
	Gdome_xml_Node *priv = (Gdome_xml_Node *)self;

	g_return_val_if_fail (priv != NULL, NULL);
	g_return_val_if_fail (GDOME_XML_IS_N (priv), NULL);
	g_return_val_if_fail (exc != NULL, NULL);

	if(GDOME_XML_IS_A (priv))
		return NULL;
	else
		return gdome_xml_n_mkref (gdome_xmlGetNext (priv->n));
}

/**
 * gdome_xml_n_nodeName:
 * @self:  Node Object ref
 * @exc:  Exception Object ref
 *
 * Returns The name of this node, depending on its type.
 * <table>
 *   <tr><td>Attr</td><td>name of attribute</td></tr>
 *   <tr><td>CDATASection</td><td>#cdata-section</td></tr>
 *   <tr><td>Comment</td><td>#comment</td></tr>
 *   <tr><td>Document</td><td>#document</td></tr>
 *   <tr><td>DocumentFragment</td><td>#document-fragment</td></tr>
 *   <tr><td>DocumentType</td><td>document type name</td></tr>
 *   <tr><td>Element</td><td>tag name</td></tr>
 *   <tr><td>Entity</td><td>entity name</td></tr>
 *   <tr><td>EntityReference</td><td>name of entity referenced</td></tr>
 *   <tr><td>Notation</td><td>notation name</td></tr>
 *   <tr><td>ProcessinInstruction</td><td>target</td></tr>
 *   <tr><td>Text</td><td>#text</td></tr>
 * </table>
 */
GdomeDOMString *
gdome_xml_n_nodeName (GdomeNode *self, GdomeException *exc)
{
	Gdome_xml_Node *priv = (Gdome_xml_Node *)self;
	GdomeDOMString *ret = NULL;

	g_return_val_if_fail (priv != NULL, NULL);
	g_return_val_if_fail (GDOME_XML_IS_N (priv), NULL);
	g_return_val_if_fail (exc != NULL, NULL);

	switch (gdome_xmlGetType (priv->n)) {
	case XML_ATTRIBUTE_NODE:
		ret = gdome_a_name((GdomeAttr *)self, exc);
		break;
	case XML_ELEMENT_NODE:
		ret = gdome_el_tagName((GdomeElement *)self, exc);
		break;
  case XML_DOCUMENT_TYPE_NODE:
  case XML_DTD_NODE:
    ret = gdome_dt_name((GdomeDocumentType *)self, exc);
    break;
  case XML_PI_NODE:
    ret = gdome_pi_target((GdomeProcessingInstruction *)self, exc);
		break;
  case XML_ENTITY_NODE:
  case XML_ENTITY_DECL: /*libxml use this instead of XML_ENTITY_NODE*/
	case XML_ENTITY_REF_NODE:
	case XML_NOTATION_NODE:
		ret = gdome_xml_str_mkref_dup((gchar *)gdome_xmlGetName (priv->n));
		break;
	case XML_CDATA_SECTION_NODE:
		ret = gdome_xml_str_mkref_dup("#cdata-section");
		break;
	case XML_COMMENT_NODE:
		ret = gdome_xml_str_mkref_dup("#comment");
		break;
	case XML_DOCUMENT_NODE:
		ret = gdome_xml_str_mkref_dup("#document");
		break;
	case XML_DOCUMENT_FRAG_NODE:
		ret = gdome_xml_str_mkref_dup("#document-fragment");
		break;
	case XML_TEXT_NODE:
		ret = gdome_xml_str_mkref_dup("#text");
		break;
	default:
		g_warning ("gdome_n_name: wrong Node Type.");
	}

	return ret;
}

/**
 * gdome_xml_n_nodeType:
 * @self:  Node Object ref
 * @exc:  Exception Object ref
 *
 * Return an integer bigger than 0 representing the type of the underlying
 * object or 0 on error.
 */
unsigned short
gdome_xml_n_nodeType (GdomeNode *self, GdomeException *exc)
{
  Gdome_xml_Node *priv = (Gdome_xml_Node *)self;
  xmlElementType type;

	g_return_val_if_fail (priv != NULL, 0);
	g_return_val_if_fail (GDOME_XML_IS_N (priv), 0);
	g_return_val_if_fail (exc != NULL, 0);

  type = gdome_xmlGetType(priv->n);
  if (type == XML_DTD_NODE)
    return XML_DOCUMENT_TYPE_NODE;
  else if (type == XML_ENTITY_DECL)
    return XML_ENTITY_NODE;
  else
    return type;
}

/**
 * gdome_xml_n_nodeValue:
 * @self:  Node Object ref
 * @exc:  Exception Object ref
 *
 * Returns The value of this node, depending on its type.
 * <table>
 *   <tr><td>Attr</td><td>values of attribute</td></tr>
 *   <tr><td>CDATASection</td><td>content of the CDATASection</td></tr>
 *   <tr><td>Comment</td><td>content of the comment</td></tr>
 *   <tr><td>Document</td><td>NULL</td></tr>
 *   <tr><td>DocumentFragment</td><td>NULL</td></tr>
 *   <tr><td>DocumentType</td><td>NULL</td></tr>
 *   <tr><td>Element</td><td>NULL</td></tr>
 *   <tr><td>Entity</td><td>NULL</td></tr>
 *   <tr><td>EntityReference</td>NULL<td></td></tr>
 *   <tr><td>Notation</td><td>NULL</td></tr>
 *   <tr><td>ProcessinInstruction</td><td>entire content excluding the target</td></tr>
 *   <tr><td>Text</td><td>content of the text node</td></tr>
 * </table>
 */
GdomeDOMString *
gdome_xml_n_nodeValue (GdomeNode *self, GdomeException *exc)
{
	Gdome_xml_Node *priv = (Gdome_xml_Node *)self;
	GdomeDOMString *ret = NULL;

	g_return_val_if_fail (priv != NULL, 0);
	g_return_val_if_fail (GDOME_XML_IS_N (priv), NULL);
	g_return_val_if_fail (exc != NULL, 0);

	switch (gdome_xmlGetType (priv->n)) {
	case XML_ATTRIBUTE_NODE:
		ret = gdome_a_value ((GdomeAttr *)self, exc);
		break;
  case XML_PI_NODE:
    ret = gdome_pi_data ((GdomeProcessingInstruction *)self, exc);
    break;
	case XML_CDATA_SECTION_NODE:
	case XML_COMMENT_NODE:
	case XML_TEXT_NODE:
		ret = gdome_xml_str_mkref_own (xmlNodeGetContent (priv->n));
		break;
	case XML_ELEMENT_NODE:
	case XML_ENTITY_NODE:
	case XML_ENTITY_REF_NODE:
	case XML_DOCUMENT_NODE:
	case XML_DOCUMENT_TYPE_NODE:
	case XML_DOCUMENT_FRAG_NODE:
  case XML_NOTATION_NODE:
  case XML_DTD_NODE:
  case XML_ENTITY_DECL:
		break;
	default:
		g_warning ("could not recognize the Node Type.");
	}

	return ret;
}

/**
 * gdome_xml_n_ownerDocument:
 * @self:  Node Objects ref
 * @exc:  Exception Object ref
 *
 * Returns the documment object (readonly) associated with this node. This is
 * also the Document object used to create new nodes. When this node is a
 * Document or a DocumentType which is not used with any Document yet, this is
 * NULL.
 */
GdomeDocument *
gdome_xml_n_ownerDocument (GdomeNode *self, GdomeException *exc)
{
	Gdome_xml_Node *priv = (Gdome_xml_Node *)self;

	g_return_val_if_fail (priv != NULL, NULL);
	g_return_val_if_fail (GDOME_XML_IS_N (priv), NULL);
	g_return_val_if_fail (exc != NULL, NULL);

	return (GdomeDocument *)gdome_xml_n_mkref ((xmlNode *)gdome_xmlGetOwner (priv->n));
}

/**
 * gdome_xml_n_parentNode:
 * @self: Node Object ref
 * @exc: Exception Object ref
 *
 * Returns the parent (readonly) of this node. All nodes, except Attr,
 * Document, DocumentFragment, Entity, and Notation may have a parent. However,
 * if a node has just been created and not yet added to the tree, or if it has
 * been removed from the tree, this is NULL.
 */
GdomeNode *
gdome_xml_n_parentNode (GdomeNode *self, GdomeException *exc)
{
	Gdome_xml_Node *priv = (Gdome_xml_Node *)self;

	g_return_val_if_fail (priv != NULL, NULL);
	g_return_val_if_fail (GDOME_XML_IS_N (priv), NULL);
	g_return_val_if_fail (exc != NULL, NULL);

	if(GDOME_XML_IS_TREE_N (priv))
		return gdome_xml_n_mkref (gdome_xmlGetParent (priv->n));
	else
		return NULL;
}

/**
 * gdome_xml_n_prefix:
 * @self:  Node Objects ref
 * @exc:  Exception Object ref
 *
 * Returns the namespace prefix of this node, or null if it is unspecified.
 * For nodes of any type other than ELEMENT_NODE and ATTRIBUTE_NODE and nodes
 * created with a DOM Level 1 method, this is always NULL.
 */
GdomeDOMString *
gdome_xml_n_prefix (GdomeNode *self, GdomeException *exc)
{
	Gdome_xml_Node *priv = (Gdome_xml_Node *)self;

	g_return_val_if_fail (priv != NULL, 0);
	g_return_val_if_fail (GDOME_XML_IS_N (priv), 0);
	g_return_val_if_fail (exc != NULL, 0);

	if((gdome_xmlGetType(priv->n) == XML_ELEMENT_NODE ||
			gdome_xmlGetType(priv->n) == XML_ATTRIBUTE_NODE) &&
		 gdome_xmlGetNs(priv->n) != NULL)
		return gdome_xml_str_mkref_dup ((gchar *)gdome_xmlGetNsPrefix (priv->n));
	else
		return NULL;
}

/**
 * gdome_xml_n_previousSibling:
 * @self:  Node Object ref
 * @exc:  Exception Object ref
 *
 * The node (readonly) immediately preceding this node. If there is no such
 * node, this returns NULL.
 */
GdomeNode *
gdome_xml_n_previousSibling (GdomeNode *self, GdomeException *exc)
{
	Gdome_xml_Node *priv = (Gdome_xml_Node *)self;

	g_return_val_if_fail (priv != NULL, NULL);
	g_return_val_if_fail (GDOME_XML_IS_N (priv), NULL);
	g_return_val_if_fail (exc != NULL, NULL);

	if(GDOME_XML_IS_A (priv))
		return NULL;
	else
		return gdome_xml_n_mkref (gdome_xmlGetPrev (priv->n));
}

/**
 * gdome_xml_n_set_nodeValue:
 * @self:  Node Object ref
 * @nodeValue:  The new value for this node.
 * @exc:  Exception Object ref
 *
 * Sets The value of this node, depending on its type.
 *
 * GDOME_NO_MODIFICATION_ALLOWED_ERR: Raised when the node is readonly.
 */
void
gdome_xml_n_set_nodeValue (GdomeNode *self, GdomeDOMString *nodeValue, GdomeException *exc)
{
	Gdome_xml_Node *priv = (Gdome_xml_Node *)self;

	g_return_if_fail (priv != NULL);
	g_return_if_fail (GDOME_XML_IS_N (priv));
	g_return_if_fail (nodeValue != NULL);
  g_return_if_fail (exc != NULL);

  if (GDOME_ISREADONLY (priv)) {
    *exc = GDOME_NO_MODIFICATION_ALLOWED_ERR;
    return;
  }

	switch (gdome_xmlGetType(priv->n)) {
	case XML_ATTRIBUTE_NODE:
		gdome_a_set_value ((GdomeAttr *)self, nodeValue, exc);
		break;
  case XML_PI_NODE:
    gdome_pi_set_data ((GdomeProcessingInstruction *)self, nodeValue, exc);
    break;
	case XML_CDATA_SECTION_NODE:
	case XML_COMMENT_NODE:
	case XML_TEXT_NODE:
		xmlNodeSetContent (priv->n, (xmlChar *)nodeValue->str);
		break;
	case XML_ELEMENT_NODE:
	case XML_ENTITY_NODE:
	case XML_ENTITY_REF_NODE:
	case XML_DOCUMENT_NODE:
	case XML_DOCUMENT_TYPE_NODE:
	case XML_DOCUMENT_FRAG_NODE:
  case XML_NOTATION_NODE:
  case XML_DTD_NODE:
		break;
	default:
		g_warning ("could not recognize the Node Type.");
  }
}

/**
 * gdome_xml_n_set_prefix:
 * @self:  Node Object ref
 * @nodeValue:  The new prefix for this node.
 * @exc:  Exception Object ref
 *
 * Sets a new nemaspace prefix for this node.
 *
 * GDOME_NO_MODIFICATION_ALLOWED_ERR: Raised when the node is readonly.
 * GDOME_NAMESPACE_ERR: Raised if the specified prefix is malformed, if the
 * namespaceURI of this node is null, if the specified prefix is "xml" and the
 * namespaceURI of this node is different from
 * "http://www.w3.org/XML/1998/namespace", if this node is an attribute and the
 * specified prefix is "xmlns" and the namespaceURI of this node is different
 * from "http://www.w3.org/2000/xmlns/", or if this node is an attribute and
 * the qualifiedName of this node is "xmlns"
 */
void
gdome_xml_n_set_prefix (GdomeNode *self, GdomeDOMString *prefix, GdomeException *exc)
{
	Gdome_xml_Node *priv = (Gdome_xml_Node *)self;
	xmlNs *ns;

	g_return_if_fail (priv != NULL);
	g_return_if_fail (GDOME_XML_IS_N (priv));
	g_return_if_fail (prefix != NULL);
	g_return_if_fail (exc != NULL);

  if (GDOME_ISREADONLY (priv)) {
    *exc = GDOME_NO_MODIFICATION_ALLOWED_ERR;
    return;
  }

  if((gdome_xmlGetType(priv->n) == XML_ELEMENT_NODE ||
			gdome_xmlGetType(priv->n) == XML_ATTRIBUTE_NODE) &&
		 gdome_xmlGetNs(priv->n) != NULL &&
     !xmlStrEqual(gdome_xmlGetNsPrefix (priv->n), (xmlChar *)prefix->str)) {

		/* tests for NAMESPACE_ERR */
    gchar *nsURI = (gchar *)gdome_xmlGetNsURI (priv->n);
		if (nsURI == NULL ||
				(!strcmp (prefix->str, "xml") &&
				 strcmp (nsURI, "http://www.w3.org/XML/1998/namespace")) ||
				(gdome_xmlGetType(priv->n) == XML_ATTRIBUTE_NODE &&
				 !strcmp (prefix->str, "xmlns") &&
				 strcmp (nsURI, "http://www.w3.org/XML/1998/namespace")) ||
				(gdome_xmlGetType(priv->n) == XML_ATTRIBUTE_NODE &&
				 !strcmp (gdome_xmlGetName(priv->n), "xmlns"))) {
			*exc = GDOME_NAMESPACE_ERR;
			return;
		}
		ns = xmlNewNs (NULL, gdome_xmlGetNsURI (priv->n), (xmlChar *)prefix->str);
		gdome_xmlSetNs(priv->n, ns);
	}
}

/**************************************
 ******* Methods implementation *******
 **************************************/

/**
 * gdome_xml_n_appendChild:
 * @self:  Node Object ref
 * @newChild:  The node to add
 * @exc:  Exception Object ref
 *
 * Adds the node newChild to the end of the list of children of this node.
 * If the newChild is already in the tree, it is first removed. If it is a
 * DocumentFragment node, the entire contents of the document fragment are
 * moved into the child list of this node
 * Returns the node added.
 *
 * GDOME_WRONG_DOCUMENT_ERR: Raised if newChild was created from a different
 * document than the one that created this node.
 * GDOME_NO_MODIFICATION_ALLOWED_ERR: Raised when the node is readonly.
 */
GdomeNode *
gdome_xml_n_appendChild (GdomeNode *self, GdomeNode *newChild, GdomeException *exc)
{
	Gdome_xml_Node *priv = (Gdome_xml_Node *)self;
	Gdome_xml_Node *new_priv = (Gdome_xml_Node *)newChild;
	xmlNode *ret;

	g_return_val_if_fail (priv != NULL, NULL);
	g_return_val_if_fail (GDOME_XML_IS_N (priv), NULL);
	g_return_val_if_fail (new_priv != NULL, NULL);
	g_return_val_if_fail (GDOME_XML_IS_N (new_priv), NULL);
	g_return_val_if_fail (exc != NULL, NULL);

	if (gdome_xmlGetOwner(new_priv->n) != gdome_xmlGetOwner(priv->n)) {
		*exc = GDOME_WRONG_DOCUMENT_ERR;
		return NULL;
	}

  if (GDOME_ISREADONLY (priv)) {
    *exc = GDOME_NO_MODIFICATION_ALLOWED_ERR;
    return NULL;
  }

  /* If the newChild is already in the tree, it is first removed. */
	if (gdome_xmlGetParent(new_priv->n) != NULL)
		gdome_xmlUnlinkChild(gdome_xmlGetParent(new_priv->n), new_priv->n);

	ret = gdome_xmlAppendChild(priv->n, new_priv->n);

	gdome_xml_mut_invoke_listener ("DOMNodeInserted",
																 newChild,
																 TRUE,
																 FALSE,
																 self,
																 NULL,
																 NULL,
																 NULL);
	
	gdome_xml_mut_invoke_listener_recursive ("DOMNodeInsertedIntoDocument",
																					 newChild,
																					 FALSE,
																					 FALSE,
																					 NULL,
																					 NULL,
																					 NULL,
																					 NULL);
	
	gdome_xml_mut_invoke_listener ("DOMSubtreeModified",
																 self,
																 TRUE,
																 FALSE,
																 newChild,
																 NULL,
																 NULL,
																 NULL);
	
	return gdome_xml_n_mkref (ret);
}

/**
 * gdome_xml_cloneNode:
 * @self:  Node Object ref
 * @deep:  If true, recursively clone the subtree under the specified node;
 *         if false, clone only the node itself (and its attributes, if it is
 *         an Element).
 * @exc:  Exception Object ref
 *
 * Returns a duplicate of this node, i.e., serves as a generic copy
 * constructor for nodes. The duplicate node has no parent; (parentNode is
 * NULL).
 */
GdomeNode *
gdome_xml_n_cloneNode (GdomeNode *self, GdomeBoolean deep, GdomeException *exc)
{
	Gdome_xml_Node *priv = (Gdome_xml_Node *)self;
	xmlNode *n = NULL;

	g_return_val_if_fail (priv != NULL, NULL);
	g_return_val_if_fail (GDOME_XML_IS_N (priv), NULL);
	g_return_val_if_fail (exc != NULL, NULL);

	n = xmlCopyNode(priv->n, deep);
	gdome_xmlSetOwner(n, gdome_xmlGetOwner(priv->n));

  return gdome_xml_n_mkref (n);
}

/**
 * gdome_xml_n_hasAttributes:
 * @self:  Node Object ref
 * @exc:  Exception Object ref
 *
 * Returns true if this node has any attributes, false otherwise.
 */
GdomeBoolean
gdome_xml_n_hasAttributes (GdomeNode *self, GdomeException *exc) {
	Gdome_xml_Node *priv = (Gdome_xml_Node *)self;

	g_return_val_if_fail (priv != NULL, 0);
	g_return_val_if_fail (GDOME_XML_IS_N (priv), 0);
	g_return_val_if_fail (exc != NULL, 0);

	if (GDOME_XML_IS_EL (priv))
		return (gdome_xmlGetAttrList (priv->n) != NULL);
	else
    return FALSE;
}

/**
 * gdome_xml_n_hasChildNodes:
 * @self:  Node Object ref
 * @exc:  Exception Object ref
 *
 * Returns true if this node has any children, false otherwise.
 */
GdomeBoolean
gdome_xml_n_hasChildNodes (GdomeNode *self, GdomeException *exc) {
	Gdome_xml_Node *priv = (Gdome_xml_Node *)self;

	g_return_val_if_fail (priv != NULL, 0);
	g_return_val_if_fail (GDOME_XML_IS_N (priv), 0);
	g_return_val_if_fail (exc != NULL, 0);

	return (gdome_xmlGetFirstChild (priv->n) != NULL);
}

/**
 * gdome_xml_n_insertBefore:
 * @self:  Node Object ref
 * @newChild:  The new node to put in the child list
 * @refChild:  the reference node, i.e., the node before which the new node
 *             must be inserted.
 * @exc:  Exception Object ref
 *
 * Inserts the node newChild before the existing child node refChild. If
 * refChild is null, insert newChild at the end of the list of children.
 * If newChild is a DocumentFragment node, all of its children are inserted,
 * in the same order, before refChild. If the newChild is already in the tree,
 * it is first removed.
 *
 * GDOME_WRONG_DOCUMENT_ERR: Raised if newChild was created from a different
 * document than the one that created this node.
 * GDOME_NOT_FOUND_ERR: Raised if oldChild is not a child of this node.
 * GDOME_NO_MODIFICATION_ALLOWED_ERR: Raised when the node is readonly.
 */
GdomeNode *
gdome_xml_n_insertBefore (GdomeNode *self, GdomeNode *newChild, GdomeNode *refChild, GdomeException *exc)
{
	Gdome_xml_Node *priv = (Gdome_xml_Node *)self;
	Gdome_xml_Node *new_priv = (Gdome_xml_Node *)newChild;
	Gdome_xml_Node *ref_priv = (Gdome_xml_Node *)refChild;
	xmlNode *ret = NULL;

	g_return_val_if_fail (priv != NULL, NULL);
	g_return_val_if_fail (GDOME_XML_IS_N (priv), NULL);
	g_return_val_if_fail (new_priv != NULL, NULL);
	g_return_val_if_fail (GDOME_XML_IS_N (new_priv), NULL);
	g_return_val_if_fail (exc != NULL, NULL);

	/* todo: occurs check for self descendant of newChild */

	if (gdome_xmlGetOwner(new_priv->n) != gdome_xmlGetOwner(priv->n)) {
		*exc = GDOME_WRONG_DOCUMENT_ERR;
		return NULL;
	}

  if (GDOME_ISREADONLY (priv)) {
    *exc = GDOME_NO_MODIFICATION_ALLOWED_ERR;
    return NULL;
  }

  /* If refChild is null, insert newChild at the end of the list of children.*/
	if(ref_priv == NULL)
		return gdome_xml_n_appendChild (self, newChild, exc);
	else
		g_return_val_if_fail (GDOME_XML_IS_N (ref_priv), NULL);

	/* If the newChild is already in the tree, it is first removed. */
	if (gdome_xmlGetParent(new_priv->n) != NULL)
		gdome_xmlUnlinkChild(gdome_xmlGetParent(new_priv->n), new_priv->n);

	if ((ret = gdome_xmlInsertBeforeChild(priv->n, new_priv->n, ref_priv->n)) == NULL) {
		*exc = GDOME_NOT_FOUND_ERR;
		return NULL;
	}

	gdome_xml_mut_invoke_listener ("DOMNodeInserted",
																 newChild,
																 TRUE,
																 FALSE,
																 self,
																 NULL,
																 NULL,
																 NULL);

	gdome_xml_mut_invoke_listener_recursive ("DOMNodeInsertedIntoDocument",
																					 newChild,
																					 FALSE,
																					 FALSE,
																					 NULL,
																					 NULL,
																					 NULL,
																					 NULL);
	
	gdome_xml_mut_invoke_listener ("DOMSubtreeModified",
																 self,
																 TRUE,
																 FALSE,
																 NULL,
																 NULL,
																 NULL,
																 NULL);

	return gdome_xml_n_mkref (ret);
}

/**
 * gdome_xml_n_isSupported:
 * @self:  Node Object ref
 * @feature:  The name of the feature to test.
 * @version:  This is the version number of the feature to test. In Level 2,
 *            version 1, this is the string "2.0". If the version is not
 *            specified, supporting any version of the feature will cause the
 *            method to return TRUE.
 * @exc:  Exception Object ref
 *
 * Tests whether the DOM implementation implements a specific feature and that
 * feature is supported by this node.
 */
GdomeBoolean
gdome_xml_n_isSupported (GdomeNode *self, GdomeDOMString *feature, GdomeDOMString *version, GdomeException *exc)
{
	Gdome_xml_Node *priv = (Gdome_xml_Node *)self;
	GdomeBoolean ret;
  GdomeDOMImplementation *domimpl;

	g_return_val_if_fail (priv != NULL, FALSE);
	g_return_val_if_fail (GDOME_XML_IS_N (priv), FALSE);
	g_return_val_if_fail (feature != NULL, FALSE);
	g_return_val_if_fail (exc != NULL, FALSE);

	if (version == NULL)
		return TRUE;

	domimpl = gdome_DOMImplementation_mkref();
	ret = gdome_DOMImplementation_hasFeature (domimpl, feature, version, exc);
	gdome_DOMImplementation_unref (domimpl, exc);

  return ret;
}

/**
 * gdome_xml_n_normalize:
 * @self:  Node Object ref
 * @exc:  Exception Object ref
 *
 * Puts all Text nodes in the full depth of the sub-tree underneath this Node,
 * including attribute nodes, into a "normal" form where only structure (e.g.,
 * elements, comments, processing instructions, CDATA sections, and entity
 * references) separates Text nodes, i.e., there are neither adjacent Text
 * nodes nor empty Text nodes.
 */
void
gdome_xml_n_normalize (GdomeNode *self, GdomeException *exc)
{
	Gdome_xml_Node *priv = (Gdome_xml_Node *)self;
	GdomeNode *child = NULL, *new_child = NULL;
	GdomeNode *next = NULL, *new_next = NULL;
	GdomeDOMString *str = NULL;
	GdomeNode *result = NULL;

	g_return_if_fail (priv != NULL);
	g_return_if_fail (GDOME_XML_IS_N (priv));
	g_return_if_fail (exc != NULL);

	/* this is a generic implementation, not optimized for speed */
	child = gdome_n_firstChild (self, exc);
	while (child != NULL) {
		unsigned short type;

		type = gdome_n_nodeType (child, exc);
		if (type == GDOME_TEXT_NODE) {
			next = gdome_n_nextSibling (child, exc);
			while (next != NULL) {
				if (gdome_n_nodeType (next, exc) != GDOME_TEXT_NODE) {
					gdome_n_unref (next, exc);
					break;
				}
				str = gdome_cd_data ((GdomeCharacterData *)next, exc);
				gdome_cd_appendData ((GdomeCharacterData *)child, str, exc);
				gdome_str_unref (str);
				new_next = gdome_n_nextSibling (next, exc);
				result = gdome_n_removeChild (self, next, exc);
				gdome_n_unref (result, exc);
				gdome_n_unref (next, exc);
				next = new_next;
			}
		} else if (type == GDOME_ELEMENT_NODE) {
			GdomeNamedNodeMap *nnm;
			unsigned long i, n;

			gdome_n_normalize (child, exc);
			nnm = gdome_n_attributes (child, exc);
			n = gdome_nnm_length (nnm, exc);
			for (i = 0; i < n; i++) {
				GdomeNode *attr;

				attr = gdome_nnm_item (nnm, i, exc);
				gdome_n_normalize (attr, exc);
				gdome_n_unref (attr, exc);
			}
			gdome_nnm_unref (nnm, exc);
		}
		new_child = gdome_n_nextSibling (child, exc);
		gdome_n_unref (child, exc);
		child = new_child;
	}
}

/**
 * gdome_xml_n_removeChild:
 * @self:  Node Object ref
 * @oldChild:  The node being removed
 * @exc:  Exception Object ref
 *
 * Removes the child node indicated by oldChild from the list of children, and
 * returns it.
 * Returns the node removed.
 *
 * GDOME_NOT_FOUND_ERR: Raised if oldChild is not a child of this node.
 * GDOME_NO_MODIFICATION_ALLOWED_ERR: Raised when the node is readonly.
 */
GdomeNode *
gdome_xml_n_removeChild (GdomeNode *self, GdomeNode *oldChild, GdomeException *exc)
{
	Gdome_xml_Node *priv = (Gdome_xml_Node *)self;
	Gdome_xml_Node *old_priv = (Gdome_xml_Node *)oldChild;
	xmlNode *ret;

	g_return_val_if_fail (priv != NULL, NULL);
	g_return_val_if_fail (GDOME_XML_IS_N (priv), NULL);
	g_return_val_if_fail (old_priv != NULL, NULL);
	g_return_val_if_fail (GDOME_XML_IS_N (old_priv), NULL);
	g_return_val_if_fail (exc != NULL, NULL);

  if (GDOME_ISREADONLY (priv)) {
    *exc = GDOME_NO_MODIFICATION_ALLOWED_ERR;
    return NULL;
  }

  if ((ret = gdome_xmlUnlinkChild(priv->n, old_priv->n)) == NULL) {
		*exc = GDOME_NOT_FOUND_ERR;
		return NULL;
	}

  gdome_xml_mut_invoke_listener ("DOMNodeRemoved",
																 oldChild,
																 TRUE,
																 FALSE,
																 self,
																 NULL,
																 NULL,
																 NULL);

	gdome_xml_mut_invoke_listener_recursive ("DOMNodeRemovedFromDocument",
																					 oldChild,
																					 FALSE,
																					 FALSE,
																					 NULL,
																					 NULL,
																					 NULL,
																					 NULL);
	gdome_xml_mut_invoke_listener ("DOMSubtreeModified",
																 self,
																 TRUE,
																 FALSE,
																 NULL,
																 NULL,
																 NULL,
																 NULL);
  return gdome_xml_n_mkref (ret);
}

/**
 * gdome_xml_n_replaceChild:
 * @self:  Node Object ref
 * @newChild:  The new node to put in the child list
 * @oldChild:  The node being replaced in the list
 * @exc:  Exception Object ref
 *
 * Replaces the child node oldChild with newChild in the list of children, and
 * returns the oldChild node. If newChild is a DocumentFragment object,
 * oldChild is replaced by all of the DocumentFragment children, which are
 * inserted in the same order.If the newChild is already in the tree, it is
 * first removed.
 * Returns the node replaced.
 *
 * GDOME_WRONG_DOCUMENT_ERR: Raised if newChild was created from a different
 * document than the one that created this node.
 * GDOME_NOT_FOUND_ERR: Raised if oldChild is not a child of this node.
 * GDOME_NO_MODIFICATION_ALLOWED_ERR: Raised when the node is readonly.
 */
GdomeNode *
gdome_xml_n_replaceChild (GdomeNode *self, GdomeNode *newChild, GdomeNode *oldChild, GdomeException *exc)
{
	g_return_val_if_fail (self != NULL, NULL);
	g_return_val_if_fail (GDOME_XML_IS_N (self), NULL);
	g_return_val_if_fail (newChild != NULL, NULL);
	g_return_val_if_fail (GDOME_XML_IS_N (newChild), NULL);
	g_return_val_if_fail (oldChild != NULL, NULL);
	g_return_val_if_fail (GDOME_XML_IS_N (oldChild), NULL);
	g_return_val_if_fail (exc != NULL, NULL);

	/* todo: occurs check for self descendant of newChild */
	/* todo: get the events right */

  if (GDOME_ISREADONLY (self)) {
    *exc = GDOME_NO_MODIFICATION_ALLOWED_ERR;
    return NULL;
  }

  if (gdome_n_insertBefore(self, newChild, oldChild, exc) == NULL)
		return NULL;

	return gdome_n_removeChild(self, oldChild, exc);
}


void
gdome_xml_n_addEventListener (GdomeNode *self, GdomeDOMString *type, 
															GdomeEventListener *listener, GdomeBoolean useCapture, 
															GdomeException *exc)
{
	Gdome_xml_Node *priv = (Gdome_xml_Node *)self;
	Gdome_xml_ListenerList *ll = NULL;
	Gdome_xml_ListenerList *temp;

	g_return_if_fail (priv != NULL);
	g_return_if_fail (GDOME_XML_IS_N (priv));
	g_return_if_fail (listener != NULL);
	g_return_if_fail (GDOME_XML_IS_EVNTL (priv));
	g_return_if_fail (type != NULL);
	g_return_if_fail (exc != NULL);

	temp = priv->ll;

	while (temp) {
		if (((Gdome_xml_EventListener *)temp->listener)->evntl->callback
				== ((Gdome_xml_EventListener *)listener)->evntl->callback &&
				strcmp (temp->type,type->str) == 0 && temp->useCapture ==
				useCapture) {
			return;
		}

		temp = temp->next;
	}

	ll = g_new (Gdome_xml_ListenerList, 1);
	ll->type = g_strdup (type->str);
	ll->listener = listener;
	ll->useCapture = useCapture;

	ll->next = priv->ll;
	priv->ll = ll;
}

void
gdome_xml_n_removeEventListener (GdomeNode *self, GdomeDOMString *type, 
																 GdomeEventListener *listener, GdomeBoolean useCapture, 
																 GdomeException *exc)
{
	Gdome_xml_Node *priv = (Gdome_xml_Node *)self;
	Gdome_xml_ListenerList *ll = NULL;

	g_return_if_fail (priv != NULL);
	g_return_if_fail (GDOME_XML_IS_N (priv));
	g_return_if_fail (listener != NULL);
	g_return_if_fail (GDOME_XML_IS_EVNTL (priv));
	g_return_if_fail (type != NULL);
	g_return_if_fail (exc != NULL);

	/* See EventTarget::removeEventListener in QUERIES */

	ll = priv->ll;
	/* todo, I'm worrying here about having to defer deletions if there
	 is an event in progress. */
}


GdomeBoolean
gdome_xml_n_dispatchEvent (GdomeNode *self, GdomeEvent *evt, GdomeException *exc)
{
	Gdome_xml_Node *priv = (Gdome_xml_Node *)self;
	int refcnt = 0;
	xmlNode *n = NULL;
	Gdome_xml_ListenerList *ll = NULL;

	if (priv == NULL)
		return 0;
	if (evt == NULL)
		return 0;
	if (exc == NULL)
		return 0;
	g_return_val_if_fail (GDOME_XML_IS_N (priv), 0);
	g_return_val_if_fail (GDOME_XML_IS_EVNT (priv), 0);


	refcnt = priv->refcnt;
	n = (xmlNode *)priv->n;
	ll = priv->ll;


	return 0;
}





/*******************************************
 **
 **         Helper functions for the 
 **         Implementation of Event
 **         Propagation
 **
 *******************************************/

static void
add_timestamp (GdomeEvent *event)
{
	Gdome_xml_EventPriv *priv = ((Gdome_xml_Event *)event)->ev;
	struct timeval tv;

	gettimeofday (&tv, NULL);

	if (priv->timestamp == 0)
		priv->timestamp = (tv.tv_sec * (long long)1000) + tv.tv_usec / 1000;
}

void
gdome_xml_n_invoke_listener (xmlNode *target,
														 const char *type,
														 GdomeEvent *event,
														 GdomeBoolean useCapture)
{
	Gdome_xml_Node *n = NULL;
	Gdome_xml_ListenerList *ll = NULL;
	GdomeException exc = 0;

  n = target->_private;

	if (n == NULL)
		return;

	add_timestamp (event);
	for (ll = n->ll; ll != NULL; ll = ll->next)
		{
			if (!strcmp (type, ll->type) && (useCapture == ll->useCapture))
				{
					gdome_evntl_handleEvent (ll->listener, event, &exc);
					/* todo: handle exceptions */
				}
		}
}

/**
 * gdome_xml_n_invoke_listener_prop: Invoke listeners with propagation.
 * @target: The actual target of the event.
 * @event: The event to propagate.
 *
 * Invokes the listeners on the event, with propagation according to
 * Section 6.2 of the 10 December, 1999 DOM2 CR.
 **/
void
gdome_xml_n_invoke_listener_prop (GdomeNode *target,
																	GdomeEvent *event)
{
	Gdome_xml_Node *priv = (Gdome_xml_Node *)target;
	xmlNode *path_stat[256];
	xmlNode **path;
	int n_path, n_path_max;
	int i;
	xmlNode *p;
	Gdome_xml_EventPriv *ev_priv = ((Gdome_xml_Event *)event)->ev;
	const GdomeBoolean do_capture = TRUE; /* always process captures
	now, this will probably
	change for speed. */

	add_timestamp (event);

	/* XXX : there is a a problem here. The routine is assuming that the
	 cast to Gdome_xml_Node is valid, but this may not be the case (eg
	 for Document or Attr nodes). I think the right thing to do is
	 dispatch based on the vtable. */
	/* note from Mathieu. I think that this is safe here with the new
		 libxml and the new structures.
		 We now have something where the cast should be valid.
		 */

	path = path_stat;

	p = priv->n;
	for (i = 0; i < sizeof(path_stat) / sizeof(path_stat[0]); i++)
		{
			if (p == NULL)
				break;
			path[i] = p;
			p = p->parent;
		}

	if (p != NULL)
		{
			n_path_max = i << 1;
			path = g_new (xmlNode *, n_path_max);
			memcpy (path, path_stat, sizeof(path_stat));
			do {
				if (i == n_path_max)
					path = g_renew (xmlNode *, path, (n_path_max <<= 1));
				i++;
				p = p->parent;
      } while (p != NULL);
		}

	n_path = i;

  ev_priv->default_prevented = FALSE;
	ev_priv->propagation_stopped = FALSE;

	if (do_capture)
		{
			ev_priv->eventPhase = GDOME_CAPTURING_PHASE;
			for (i = n_path - 1; i > 0; i--)
				{
					/* Set currentTarget on the event before invoking listeners */
					ev_priv->currentTarget = path[i];

					gdome_xml_n_invoke_listener (path[i], ev_priv->type, event, TRUE);
					if (ev_priv->propagation_stopped)
						break;
				}
		}

	if (!ev_priv->propagation_stopped)
		{
			ev_priv->eventPhase = GDOME_AT_TARGET;
			ev_priv->currentTarget = path[0];
			gdome_xml_n_invoke_listener (path[0], ev_priv->type, event, FALSE);

			if (!ev_priv->propagation_stopped && ev_priv->bubbles)
				{
					ev_priv->eventPhase = GDOME_BUBBLING_PHASE;
					for (i = 1; i < n_path; i++)
						{
							/* Set currentTarget on the event before invoking listeners */
							ev_priv->currentTarget = path[i];

							gdome_xml_n_invoke_listener (path[i], ev_priv->type, event,
																					 FALSE);
							if (ev_priv->propagation_stopped)
								break;
						}
				}
		}

	if (path != path_stat)
		g_free (path);
}

/* This is for mutation events only, currently. */
void
gdome_xml_mut_invoke_listener (const char *type,
																 GdomeNode *target,
																 GdomeBoolean canBubble,
																 GdomeBoolean cancelable,
																 GdomeNode *relatedNode,
																 const char *prevValue,
																 const char *newValue,
																 const char *attrName)
{
	Gdome_xml_MutationEventPriv mev;
	Gdome_xml_MutationEvent event;
	GdomeDOMString prevValue_str;
	GdomeDOMString newValue_str;
	GdomeDOMString attrName_str;

	prevValue_str.unref = gdome_xml_str_unref_nop;
	prevValue_str.str = (char *)prevValue;

	newValue_str.unref = gdome_xml_str_unref_nop;
	newValue_str.str = (char *)newValue;

	attrName_str.unref = gdome_xml_str_unref_nop;
	attrName_str.str = (char *)attrName;

	event.super.vtab = (GdomeMutationEventVtab *)&gdome_xml_mevnt_vtab;
	event.refcnt = 1;
	event.mev = &mev;

	mev.ev.type = (char *)type;
	mev.ev.target = target;
	mev.ev.bubbles = canBubble;
	mev.ev.cancelable = cancelable;

	mev.ev.propagation_stopped = FALSE;
	mev.ev.timestamp = 0;
	
	if (prevValue != NULL)
		mev.prevValue = &prevValue_str;
	else
		mev.prevValue = NULL;

	if (newValue != NULL)
		mev.newValue = &newValue_str;
	else
		mev.newValue = NULL;

	if (attrName != NULL)
		mev.attrName = &attrName_str;
	else
		mev.attrName = NULL;

	mev.relatedNode = relatedNode;

	gdome_xml_n_invoke_listener_prop (target, (GdomeEvent *)&event);
}

/* Invoke a node's event listener and all its children's event listeners */
void
gdome_xml_mut_invoke_listener_recursive (const char *type,
																				 GdomeNode *target,
																				 GdomeBoolean canBubble,
																				 GdomeBoolean cancelable,
																				 GdomeNode *relatedNode,
																				 const char *prevValue,
																				 const char *newValue,
																				 const char *attrName)
{
				xmlNode *n;

				n = gdome_xml_n_get_xmlNode (target);

				/* hrm, recursive propagation does not work */
#if 0
				while (n) {
								if (n->children) {
												gdome_xml_mut_invoke_listener_recursive (type,
																																 target,
																																 canBubble,
																																 cancelable,
																																 relatedNode,
																																 prevValue,
																																 newValue,
																																 attrName);
								}
								n = n->next;
				}
#endif
}














